import os
import tempfile
import shutil
import plyvel, json

import pytest

from pypeline import DB
from pypeline._version import schema_version


@pytest.fixture
def db_dir(request):
    path = tempfile.mkdtemp()

    def finalize():
        shutil.rmtree(path)

    request.addfinalizer(finalize)
    return path

@pytest.fixture
def db(request):
    path = tempfile.mkdtemp()
    db = DB(path, create_if_missing=True)

    def finalize():
        db.close()
        shutil.rmtree(path)

    request.addfinalizer(finalize)
    return db

@pytest.fixture
def collection(request):
    path = tempfile.mkdtemp()
    db = DB(path, create_if_missing=True)

    collection = db.collection('test')

    def finalize():
        db.close()
        shutil.rmtree(path)

    request.addfinalizer(finalize)
    return collection


def test_db_creation(db_dir):
    test_db = DB(db_dir, create_if_missing=True)
    assert os.path.isdir(db_dir) == True
    assert isinstance(test_db, DB) == True

def test_db_schema(db_dir):
    test_db = DB(db_dir, create_if_missing=True)
    test_db.close()

    test_levelDB = plyvel.DB(db_dir)
    assert json.loads(test_levelDB.get(b'pypeline-schema-version').decode()) == schema_version

def test_collection_creation(db):
    collection = db.collection('test')

    # Collection names cannot contain the reserved '!!' sequence
    with pytest.raises(ValueError):
        db.collection('a !! collection') 
    with pytest.raises(ValueError):
        db.collection('nonexistant-collection', create_if_missing=False)
    with pytest.raises(ValueError):
        db.collection('test', error_if_exists=True)

def test_collection_reset(db):
    collection = db.collection('test')

def test_list_collections(db):
    collections = ['test', 'test1', 'test2']
    for collection in collections:
        db.collection(collection)

    assert db.collections() == collections

def test_append(collection):
    collection.append([1,2])
    collection.append({'a': 'b'})

    assert collection[0] == [1,2]

    assert [c for c in collection] == [[1,2], {'a': 'b'}]

def test_append_all(db):
    c1 = db.collection('test1')
    c2 = db.collection('test2')
    c1.append(1)
    c2.append(2)
    c2.append(3)

    c1.append_all(c2)
    c1.append_all([4,5])
    assert [instance for instance in c1] == [1,2,3,4,5]


def test_collection_length(db):
    collection = db.collection('test')
    collection.append([1,2,3])
    collection.append([4,5,6])

    assert len(collection) == 2

def test_collection_random_access(collection):
    collection.append([1,2])
    collection.append({'a': 'b'})
    collection.append([3,4])

    assert collection[0] == [1,2]
    assert collection[1] == {'a': 'b'}
    assert collection[2] == [3,4]

    with pytest.raises(IndexError):
        assert collection[3] == [3,4]

def test_collection_slicing(collection):
    collection.append([1,2])
    collection.append({'a': 'b'})
    collection.append([3,4])

    assert collection[0:2] == [[1,2],{'a':'b'}]

def test_collection_random_access_assign(collection):
    collection.append([1,2])
    collection.append([3,4])
    collection[1] = [5,6]

    assert collection[1] == [5,6]

    with pytest.raises(IndexError):
        collection[3] = [3,4]

def test_collection_delete(collection):
    collection.append("a")
    collection.append("b")

    collection.delete(0)
    assert len(collection) == 1
    assert collection[0] == "b"

    collection.refresh()
    assert len(collection) == 1
    assert collection[0] == "b"


def test_collection_refresh(db):
    c1 = db.collection('test')
    c2 = db.collection('test')

    c1.append([1,2])

    assert len(c1) == 1
    assert len(c2) == 1
    assert c1[0] == [1,2]
    assert c2[0] == [1,2]

def test_collection_delete(db):
    c1 = db.collection('test')
    c1.append(1)

    db.delete('test')

    with pytest.raises(ValueError):
        c2 = db.collection('test', create_if_missing=False)

    c2 = db.collection('test')
    assert len(c2) == 0

def test_collection_shallow_copy(db):
    c1 = db.collection('test')
    c2 = c1

    c1.append([1,2])

    assert len(c1) == 1
    assert len(c2) == 1
    assert c1[0] == [1,2]
    assert c2[0] == [1,2]

def test_collection_deep_copy(db):
    c1 = db.collection('c1')
    c1.append(0)
    c2 = db.copy_collection('c1', 'c2')
    c1.append(1)
    c2.append(2)
    assert [instance for instance in c1] == [0, 1]
    assert [instance for instance in c2] == [0, 2]

    c1.append_all(range(3,10))
    c3 = db.copy_collection('c1', 'c3', start=2, end=5)
    assert c3 == db.collection('c3')
    assert [instance for instance in c3] == [3,4,5]

    with pytest.raises(ValueError):
        db.copy_collection('c1', 'c5', create_if_missing=False)
    with pytest.raises(ValueError):
        db.copy_collection('c1', 'c2', error_if_exists=True)

def test_database_reloading(db_dir):
    test_db = DB(db_dir, create_if_missing=True)
    c1 = test_db.collection('test')
    c1.append(5)
    c1.append(6)

    c1.delete(0)

    test_db.close()

    test_db2 = DB(db_dir)
    assert test_db2.collection('test')[0] == 6
    assert len(test_db2.collection('test')) == 1

def test_maps(db):
    c1 = db.collection('c1')
    c1.append_all([1,2,3])
    c1.map(lambda x: x+1, None)
    assert [instance for instance in c1] == [2,3,4]
    c1.map(lambda x: x+1, 'c2')
    assert [instance for instance in c1] == [2,3,4]
    assert [instance for instance in db.collection('c2')] == [3,4,5]
    with pytest.raises(ValueError):
        c1.map(lambda x: x+1, 'c2', error_if_exists=True)
    with pytest.raises(ValueError):
        c1.map(lambda x: x+1, 'c5', create_if_missing=False)

def test_filters(db):
    c1 = db.collection('c1')
    c1.append_all(range(1,10))
    c1.filter(lambda x: x > 3, None)
    assert [instance for instance in c1] == [4,5,6,7,8,9]
    c1.filter(lambda x: x < 7, 'c2')
    assert [instance for instance in db.collection('c2', create_if_missing=False)] == [4,5,6]

    with pytest.raises(ValueError):
        c1.filter(lambda x: x > 1, 'c2', error_if_exists=True)
    with pytest.raises(ValueError):
        c1.filter(lambda x: x > 1, 'c5', create_if_missing=False)

def test_reduce(db):
    c1 = db.collection('c1')
    c1.append_all([1,2,3])
    c2 = c1.reduce(lambda x,y: x+y, 'c2')
    assert len(c2) == 1
    assert c2[0] == 6

    c1.reduce(lambda x,y: x+y, 'c2', initializer=5)
    assert len(c2) == 1
    assert c2[0] == 11
    with pytest.raises(ValueError):
        c1.reduce(lambda x,y: x+y, 'c2', error_if_exists=True)
    with pytest.raises(ValueError):
        c1.reduce(lambda x,y: x+y, 'c5', create_if_missing=False)


def test_random_subset(db):
    c1 = db.collection('c1')
    c1.append_all(range(0,10))
    c1.random_subset(8, None)
    assert len(c1) == 8
    c1.refresh()
    assert len(c1) == 8
    c2 = c1.random_subset(5, 'c2')
    assert len(c2) == 5

    with pytest.raises(ValueError):
        c1.random_subset(5, 'c2', error_if_exists=True)
    with pytest.raises(ValueError):
        c1.random_subset(5, 'c5', create_if_missing=False)


import json
import copy
import random
from functools import reduce
import plyvel
from ._version import schema_version

class DB:
    """
    The pypeline LevelDB database.  This class contains collections and
    provides some system-level organization.

    All arguments beyond the database path are passed directly to the
    underlying plyvel DB constructor. More details can be found at
    https://plyvel.readthedocs.org/en/latest/api.html#DB 

    Arguments:
    `database_path` -- The path to the folder for database storage
    """

    def __init__(self, database_path, **kwargs):         
        self.db=plyvel.DB(database_path, **kwargs)

        self.collections_set=self.db.prefixed_db(b'collections/')
        self.collection_items_set=self.db.prefixed_db(b'collection-items/')
        self.collections_cache = {}

        try:
            self.schema_version = decode(self.db.get(b'pypeline-schema-version'))
        except:
            self.schema_version = schema_version
            self.db.put(b'pypeline-schema-version', encode(schema_version))

    def collection(self, collection_name, reset_collection=False, 
        create_if_missing=True, error_if_exists=False):
        """
        Returns the collection stored at `collection_name`, or creates it if it doesn't exist.

        | Arguments:
        | ``collection_name`` -- the name of the collection to return

        | Keyword arguments:
        | ``reset_collection`` -- when True any existant data in the collection is deleted before it is returned
        | ``create_if_missing`` -- when False a ValueError is raised if the collection doesn't exist
        | ``error_if_exists`` -- When True a ValueError is raised if the collection already exists 
        """

        if collection_name not in self.collections_cache:
            data = self.collections_set.get(collection_name.encode())
            if data == None:
                if create_if_missing == False:
                    raise ValueError("Collection '{0}' does not exist".format(collection_name))
                self.collections_set.put(collection_name.encode(), b'true')
            self.collections_cache[collection_name] = Collection(self, self.collection_items_set, collection_name)
        elif error_if_exists:
            raise ValueError("Collection '{0}' already exists".format(collection_name))

        if reset_collection:
            self.collections_cache[collection_name].delete_all()

        return self.collections_cache[collection_name]

    def collections(self):
        """
        Returns a list of keys of collections contained in the database.
        """
        
        return [name.decode() for name in self.collections_set.iterator(include_value=False)]

    def copy_collection(self, old_collection, new_collection, start=None, end=None, **kwargs):
        """
        Copies all instances in the old_collection into the new_collection

        | Arguments:
        | ``old_collection`` -- The string name of the old collection
        | ``new_collection`` -- The string name of the new collection

        | Keyword arguments:
        | ``start`` -- (Optional) The index to begin copying from old_collection
        | ``end`` -- (Optional) The index to end copying from old_collection
        | ``create_if_missing`` -- when False a ValueError is raised if the new collection doesn't exist (default: True)
        | ``error_if_exists`` -- When True a ValueError is raised if the collection already exists (default: False)
        """
        old = self.collection(old_collection, create_if_missing=False)
        new = self.collection(new_collection, reset_collection=True, **kwargs)
        new.append_all(old.iterator(start, end))

        return new

    def delete(self, collection_name):
        """
        Deletes a collection.

        | Arguments:
        | ``collection_name`` -- the name of the collection to delete
        """

        self.collection(collection_name).delete_all()
        self.collections_set.delete(collection_name.encode())
        del self.collections_cache[collection_name]

    def close(self):
        """Closes the database."""
        self.db.close()

    def open(self):
        """Opens the database."""
        self.db.open()

class Collection:
    """
    A collection of records stored in a database

    This class should never be instantiated directly.  Use the ``DB.collection()`` method instead
    """
    def __init__(self, database, items_set, name):
        if '!!' in name:
            raise ValueError("Disallowed character sequence '!!' in collection name")

        self.name = name
        self.db = items_set.prefixed_db(name.encode()+b'!!')
        self.parent_db = database

        self.refresh()

    def append(self, record):
        """
        Appends a single record.

        | Arguments:
        | ``record`` -- Any JSON-serializable python object (dicts, lists, ints, strings, etc.)
        """
        self.last_index += 1
        key = str(self.last_index).encode()
        self.db.put(key, encode(record))
        self.keys.append(key)

    def refresh(self):
        """
        Reloads the collection from the database.
        """
        self.keys = []
        for key in self.db.iterator(include_value=False):
            self.keys.append(key)

        if len(self.keys) > 0:
            self.last_index = int(self.keys[-1])
        else:
            self.last_index = 0

    def delete(self, index):
        """
        Deletes an item from the collection.

        | Arguments:
        | ``index`` -- Index of the item to be deleted.
        """
        self.db.delete(self.keys[index])
        self.keys.pop(index)

    def delete_all(self):
        """Deletes all items in the collection"""

        for key in self.keys:
            self.db.delete(key)
        self.keys = []
        self.last_index = 0

    def append_all(self, iterable):
        """Appends every item in the iterable to the collection"""

        for instance in iterable:
            self.append(instance)

    def map(self, function, new_collection, **kwargs):
        """
        Maps a collection to a new collection with a provided function.

        | Arguments:
        | ``function`` -- The function used for mapping.
        | ``new_collection`` -- The name of the collection to insert the new values into.  
            Any existing values will be deleted.
            If ``None``, values are mapped to the same collection.

        | Keyword arguments:
        | ``create_if_missing`` -- when False a ValueError is raised if the new collection doesn't exist
        | ``error_if_exists`` -- When True a ValueError is raised if the new collection already exists 
        """
        collection = None
        if new_collection in [None, self.name]:
            collection = self
            for key in self.keys:
                new_value = function(decode(self.db.get(key)))
                self.db.put(key, encode(new_value))
        else:
            collection = self.parent_db.collection(new_collection, reset_collection=True, **kwargs)
            for instance in self:
                collection.append(function(instance))

        return collection

    def filter(self, function, new_collection, **kwargs):
        """
        Filters a collection into a new collection with a given function.

        | Arguments:
        | ``function`` -- The function used for filtering.
        | ``new_collection`` -- The name of the collection to insert the new values into.  
            Any existing values will be deleted.
            If ``None``, values are filtered in the same collection.

        | Keyword arguments:
        | ``create_if_missing`` -- when False a ValueError is raised if the new collection doesn't exist
        | ``error_if_exists`` -- When True a ValueError is raised if the new collection already exists 
        """
        collection = None
        if new_collection in [None, self.name]:
            collection = self
            new_keys = []
            for key in self.keys:
                if function(decode(self.db.get(key))):
                    new_keys.append(key)
                else:
                    self.db.delete(key)
            self.keys = new_keys

        else:
            collection = self.parent_db.collection(new_collection, reset_collection=True, **kwargs)
            for instance in self:
                if function(instance):
                    collection.append(instance)

        return collection

    def reduce(self, function, new_collection, initializer=None, **kwargs):
        """
        Reduces a collection into a new collection with a given function.

        | Arguments:
        | ``function`` -- The function used for reducing.
        | ``new_collection`` -- The name of the collection to insert the new value into.  
            Any existing values will be deleted.
            If ``None``, the current collection is replaced with the reduction output.

        | Keyword arguments:
        | ``create_if_missing`` -- when False a ValueError is raised if the new collection doesn't exist
        | ``error_if_exists`` -- When True a ValueError is raised if the new collection already exists 
        """
        reduced = None
        if initializer != None:
            reduced = reduce(function, self.iterator(), initializer)
        else:
            reduced = reduce(function, self.iterator())

        collection = None
        if new_collection in [None, self.name]:
            collection = self
        else:
            collection = self.parent_db.collection(new_collection, reset_collection=True, **kwargs)
        
        collection.append(reduced)
        return collection

    def random_subset(self, number, new_collection, **kwargs):
        """
        Produces a random subset of a given collection and inserts it into a new collection.

        | Arguments:
        | ``new_collection`` -- The name of the collection to insert the new values into.  
            Any existing values will be deleted.
            If `None`, the subset is stored to the current collection.

        | Keyword arguments:
        | ``create_if_missing`` -- when False a ValueError is raised if the new collection doesn't exist
        | ``error_if_exists`` -- When True a ValueError is raised if the new collection already exists 
        """
        collection = None
        if new_collection in [None, self.name]:
            collection = self
            random.shuffle(self.keys)
            for key in self.keys[number:]:
                self.db.delete(key)
            self.keys = self.keys[:number]
            list.sort(self.keys)

        else:
            new_keys = copy.copy(self.keys)
            random.shuffle(new_keys)
            new_keys = new_keys[:number]
            list.sort(new_keys)
            collection = self.parent_db.collection(new_collection, **kwargs)
            collection.delete_all()
            for key in new_keys:
                collection.append(decode(self.db.get(key)))

        return collection


    def iterator(self, start=None, end=None):
        """Returns a collection iterator"""

        return Iterator(self, start, end)

    def __iter__(self):
        return self.iterator()

    def __getitem__(self, key):
        if isinstance(key, slice):
            return [decode(self.db.get(key)) for key in self.keys[key]]
        else:
            return decode(self.db.get(self.keys[key]))

    def __setitem__(self, key, value):
        self.db.put(self.keys[key], encode(value))

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.name)

    def __len__(self):
        return len(self.keys)

class Iterator:
    def __init__(self, collection, start=None, end=None):
        self.key_iterator = iter(collection.keys[start:end])
        self.collection = collection

    def __iter__(self):
        return self

    def next(self):
        return decode(self.collection.db.get(self.key_iterator.next()))

    def __next__(self):
        return decode(self.collection.db.get(self.key_iterator.__next__()))

def encode(obj):
    return json.dumps(obj).encode()

def decode(string):
    return json.loads(string.decode())



import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

with open(os.path.join(os.path.dirname(__file__), 'pypeline', '_version.py')) as f:
    exec(f.read(), globals(), locals())


long_description = ''
if os.path.exists('README.txt'):
    long_description = open('README.txt').read()

setup(
    name = "pypeline-db",
    version = __version__,
    author = "Kyle Corbitt",
    author_email = "kyle@corbt.com",
    description = "A database for processing and storing datasets based on LevelDB",
    license = "MIT",
    keywords = "levelDB big data data science",
    url = "https://github.com/kcorbitt/pypeline",
    packages=['pypeline'],
    long_description=long_description,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "Operating System :: POSIX",
        "Topic :: Utilities",
        "Topic :: Database",
        "Topic :: Scientific/Engineering",
    ],
    install_requires=['plyvel']
)
